from datetime import (
    date,
    timedelta,
)
from re import findall
from typing import (
    List,
    Optional,
)

from .calendar import (
    Agenda,
    Event,
)
from .config import Config
from .confscr import Conf
from .culdav import Culdav
from .helpscr import Help
from .i18n import _
from .inputs import Input
from .ui import UI


class Culendar:
    WE_DAYS = (6, 7)  # Saturday, Sunday

    conf: Config
    ui: UI
    cal: Agenda
    todo: Agenda
    caldav: Culdav
    calweek: List[Agenda]
    count: int  # repetitions of the next command
    prevehl: int
    ehl: int  # highlighted event of current day
    todohl: int  # highlighted todo, if any, index starts at 1, 0 means focus on calendar
    _prevday: date
    _day: date

    def __init__(self, screen):
        self.conf = Config()
        self.ui = UI(self, self.conf, screen)
        self.cal, self.todo = Agenda.load(self.conf.datafile, self.conf.todofile)
        self.culdav = Culdav(self.conf.caldav, self.conf.datapath, self.ui)
        self.culdav.initialize()  # get cals and associated tags
        self.culdav.sync()  # update and redefine _events
        self.calweek = []
        self.count = 0
        self.prevehl = self.ehl = self.todohl = 0
        self.day = None  # current day
        self.ui.update()

    @property
    def day(self):
        return self._day

    @day.setter
    def day(self, day: Optional[date]):
        if day is None:
            self._prevday = date(1, 1, 1)
            self._day = date.today()
        else:
            self._prevday = self._day
            self.changehl(-self.ehl)  # resets highlight
            self._day = day
        self.avoid_WE(self.previous_day <= self.day)
        if not self.is_hl_same_week:
            # if it's a new week, recompute calweek
            self.update_calweek()

    @property
    def daynb(self):
        return self.ui.daynb

    @property
    def previous_day(self):
        return self._prevday

    def avoid_WE(self, forward=True):
        if self.conf.WE == 0:
            while self._day.isoweekday() in self.WE_DAYS:
                if forward:
                    self._day = date.fromordinal(self._day.toordinal() + 1)
                else:
                    self._day = date.fromordinal(self._day.toordinal() - 1)

    def update_calweek(self, autosync: Optional[bool] = None):
        """extraction of the agenda of all interesting days"""
        self.calweek = Agenda.extract_week(self.cal, self.culdav, self.day, self.ui, self.conf.autosync if autosync is None else autosync)

    def set_same_day(self):
        self._prevday = self._day

    @property
    def is_hl_same_day(self):
        return self.previous_day == self.day

    @property
    def is_hl_hidden_WE(self):
        iprevday = self.previous_day.isoweekday() - 1
        return iprevday >= self.daynb

    @property
    def is_hl_same_week(self):
        """same year and same isoweek"""
        return self.day.isocalendar()[:2] == self.previous_day.isocalendar()[:2]

    def toggle_WE(self):
        self.conf.WE = 1 - self.conf.WE
        self.avoid_WE(False)
        self.ui.update()

    def run_loop(self):
        key = ""
        while key not in self.conf.keys['quit']:
            key = self.ui.getkey()
            # repeat action at least 1, or the entered count value
            repetitions = max(1, self.count)
            # changing day or week or year
            if key in self.conf.keys['prevday']:
                self.addday(-repetitions)
            if key in self.conf.keys['nextday']:
                self.addday(+repetitions)
            if key in self.conf.keys['nextweek']:
                self.addday(+7 * repetitions)
            if key in self.conf.keys['prevweek']:
                self.addday(-7 * repetitions)
            if key in self.conf.keys['nextmonth']:
                self.addmonth(+repetitions)
            if key in self.conf.keys['prevmonth']:
                self.addmonth(-repetitions)
            if key in self.conf.keys['nextyear']:
                self.addyear(+repetitions)
            if key in self.conf.keys['prevyear']:
                self.addyear(-repetitions)
            if key in self.conf.keys['today']:
                self.changeday(date.today())
            if key in self.conf.keys['setday']:
                self.changeday()
            if key in self.conf.keys['startweek']:
                self.addday(-self.day.isoweekday() + 1)
            if key in self.conf.keys['endweek']:
                self.addday(-self.day.isoweekday() + self.daynb)
            # change highlighted event of current day
            if key in self.conf.keys['nextevent']:
                self.set_same_day()
                self.changehl(+repetitions)
                self.ui.draw_hl()
            if key in self.conf.keys['prevevent']:
                self.set_same_day()
                self.changehl(-repetitions)
                self.ui.draw_hl()
            # delete, edit, add (highlighted) event (if existing)
            if key in self.conf.keys['delevent']:
                self.del_event()
            if key in self.conf.keys['editevent']:
                self.edit_event()
            if key in self.conf.keys['addevent']:
                self.add_event()
            if key in self.conf.keys['tagevent']:
                self.tag_event()
            if key in self.conf.keys['copyevent']:
                self.copy_event(repetitions)
            if key in self.conf.keys['minusshifthour']:
                self.shift_event(-repetitions)
            if key in self.conf.keys['minusshiftday']:
                self.shift_event(-repetitions * 24)
            if key in self.conf.keys['shifthour']:
                self.shift_event(+repetitions)
            if key in self.conf.keys['shiftday']:
                self.shift_event(+repetitions * 24)
            # toggle week-end display
            if key in self.conf.keys['toggleWE']:
                self.toggle_WE()
            # toggle Todo bar focus (appear/disappear)
            if key in self.conf.keys['toggletodo']:
                self.toggle_TODO()
            if key in self.conf.keys['redraw']:
                self.ui.clear_cal()
                self.ui.draw_cal()
            if key in self.conf.keys['sync']:
                self.sync_caldav()
            if key in self.conf.keys['help']:
                self.ui.do_update_on_resize = False
                help = Help(self.ui, self.conf)
                help.help_screen()
                self.ui.update()
                self.ui.do_update_on_resize = True
            if key in self.conf.keys['save']:
                self.ui.inform(_("Saving configuration and calendar…"))
                self.conf.save()
                self.save()
                self.ui.inform(_("Configuration and calendar saved!"))
            if key in self.conf.keys['setconfig']:
                self.ui.do_update_on_resize = False
                conf = Conf(self.ui, self.conf, self.culdav)
                # real copy to keep a backup
                old_caldav_conf = [conf for conf in self.conf.caldav]
                # conf_screen returns a possibly modified conf
                self.conf = conf.conf_screen()
                self.ui.do_update_on_resize = True
                if old_caldav_conf != self.conf.caldav:
                    # FIXME should be updated with self.conf.caldav instead
                    self.culdav.update(old_caldav_conf)
                # redraw, works even in case of KEY_RESIZE inside help
                self.ui.update()
            if key in self.conf.keys['import']:
                self.importcal()
            if key in self.conf.keys['export']:
                self.exportcal()
            if key == "KEY_MOUSE":
                self.clicked()
            # define the buffer and the repetition of actions
            try:  # is the key an integer?
                # type 1 then 2, get 12
                self.count = self.count * 10 + int(key)
            except ValueError:  # not a number, back to zero
                self.count = 0
        # autosave on quit
        if self.conf.autosave:
            self.conf.save()
            self.save()

    def toggle_TODO(self):
        # if todohl > 1, Todo bar is in highlight mode, remove it
        if self.todohl > 0:
            self.conf.todo = False
            self.todohl = 0
            self.ui.update()  # recompute everything
        elif self.conf.todo:  # Todo bar is unhighlighted
            self.todohl = 1
            self.ui.draw_day()  # remove the day hl
            self.ui.draw_header()  # remove the week hl
            self.ui.draw_hl()  # hl the Todo item
        else:  # make appear the Todo bar
            self.conf.todo = True
            self.todohl = 1
            self.ui.update()  # recompute everything

    def addyear(self, shift):
        try:
            daytmp = self.day.replace(year=self.day.year + shift)
            delta = daytmp - self.day
            self.addday(delta.days)  # use the proper function
        except Exception:
            # we can't, that's a daytmp is out of range for month
            # ie, from 19 February to 31 February
            self.addday(shift * 365)  # fallback function

    def addmonth(self, shift):
        newmonth = (self.day.month - 1 + shift) % 12 + 1
        try:
            if abs(self.day.month - newmonth) != abs(shift):  # year changed
                self.addyear(int(shift / abs(shift)))  # sign(shift)
            daytmp = self.day.replace(month=newmonth)
            delta = daytmp - self.day
            self.addday(delta.days)  # use the proper function
        except Exception:
            # we can't, that's a day is out of range for month
            # ie, from 31 January to 31 February
            self.addday(shift * 30)  # fallback function

    def addday(self, shift):
        # change day: autoswitch from todo bar if any
        if self.todohl:
            self.todohl = 0
            self.ui.draw_header()  # hl the weekbar
            self.ui.draw_todo()  # un-hl the todo bar
        # set a new day
        self.day = self.day + timedelta(days=shift)
        # update the highlight, which remembers old values of day and ehl
        self.changehl(0)
        self.ui.draw_week()

    def changehl(self, inc):
        if self.todohl:  # Todo bar is highlighted
            maxhl = len(self.todo.events)
            self.todohl = (self.todohl + inc) % (maxhl + 1)
            if self.todohl == 0:
                self.todohl = 1  # todo hl starts at index 1
        else:
            self.prevehl = self.ehl
            iday = self.day.isoweekday() - 1
            maxhl = len(self.calweek[iday].events)
            if maxhl > 0:
                self.ehl = (self.ehl + inc) % maxhl
            else:
                self.ehl = 0

    def changeday(self, day=None):
        self.ui.clear_footer()
        if not day:
            day = self.ask_day()
            if day == "CUL_CANCEL":
                return  # cancel the change of day
        self.day = day
        self.avoid_WE(False)
        self.ui.draw_week()

    def sync_caldav(self):
        # update caldavs only
        self.culdav.sync()
        # redraw if something changed
        self.ui.clear_cal()
        self.ui.draw_cal()

###############################################################################
#       event functions
###############################################################################

    def del_event(self):
        if self.todohl > 0:
            self.del_todo()
            return
        inform = 0
        # select highlighted event if existing
        iday = self.day.isoweekday() - 1
        if len(self.calweek[iday].events) > self.ehl:
            event = self.calweek[iday].events[self.ehl]
            # clear the footer event
            self.ui.clear_footer()
            if event.caldav:
                if event.caldav == "webcal":
                    # webcals are read-only
                    msg = _("Error: can't delete webcal event or offline copy")
                    self.ui.text_in_footer(msg, _("Press any key to continue"))
                    self.ui.getkey()
                    self.ui.clear_footer()
                else:
                    # delete from local cal in order to avoid syncing caldav to update events
                    self.culdav.del_event(event)
            else:
                # delete it from the cal
                self.cal.del_event(event)
            self.update_calweek(False)
            # redraw day without the deleted event
            self.changehl(0)  # to update the highlighted event
            self.ui.draw_week()  # week instead of day for multidays event
            if inform:
                self.ui.inform(msg + event.url)

    def del_todo(self):
        # reminder: todo_hl starts at index 1
        self.todo.del_event(self.todo.events[self.todohl - 1])
        self.changehl(0)  # to update the highlighted event
        self.ui.clear_todo()
        self.ui.draw_todo()

    def edit_event(self):
        if self.todohl > 0:
            self.edit_todo()
            return
        # select highlighted event if existing
        iday = self.day.isoweekday() - 1
        if len(self.calweek[iday].events) > self.ehl:
            event = self.calweek[iday].events[self.ehl].original
            # we can't edit webcal events
            if event.caldav == "webcal":
                # webcals are read-only
                self.ui.clear_footer()
                msg = _("Error: can't delete webcal event")
                self.ui.text_in_footer(msg, _("Press any key to continue"))
                self.ui.getkey()
                self.ui.clear_footer()
                return  # exit here
            # the footer will be useful, clear it
            self.ui.clear_footer()
            # ask question
            label = _("Edit: (1) [s]tart time, (2) [e]nd time, (3) [d]escription, (4) [p]lace")
            keys = findall(r'\[(.)\]', label)
            self.ui.text_in_footer(label, '')
            s = self.ui.getkey()
            self.ui.clear_footer()
            if s in ('1', keys[0]):
                ed = event.date + timedelta(seconds=event.duration)
                sd = self.ask_starttime()
                if sd == "CUL_CANCEL":
                    return   # cancel edit event
                if sd > ed:  # shift the event to the new starting hour
                    d = event.duration
                else:  # keep the same end hour
                    # recompute duration
                    d = (ed - sd).total_seconds()
                event.date = sd
                event.duration = d
            elif s in ('2', keys[1]):
                ed = self.ask_endtime(event.date)
                if ed == "CUL_CANCEL":
                    return  # cancel edit event
                d = (ed - event.date).total_seconds()
                event.duration = d
            elif s in ('3', keys[2]):
                desc, place = self.ask_description(event)
                if desc == "CUL_CANCEL":
                    return  # canceled edition
                event.summary = desc
                if place:  # non empty
                    event.place = place
            elif s in ('4', keys[3]):
                place = self.ask_place(event)
                if place == "CUL_CANCEL":
                    return  # canceled edition
                event.place = place
            else:
                return
            # sort the events
            self.cal.sort()
            self.update_calweek(False)
            if event.caldav:  # update the corresponding caldav
                self.culdav.del_event(event)
                self.culdav.add_event(event, event.caldav)
            # redraw full week for multidays events
            self.ui.draw_week()

    def edit_todo(self):
        # the footer will be useful, clear it
        self.ui.clear_footer()
        if len(self.todo.events) < self.todohl - 1:
            return
        todo = self.todo.events[self.todohl - 1]
        question = _("Edit: (1) [i]tem, (2) [d]ate, (3) [r]emove date")
        keys = findall(r'\[(.)\]', question)
        self.ui.text_in_footer(question, '')
        s = self.getkey()
        if s == "KEY_RESIZE":
            self.update()
            return
        if s in ('1', keys[0]):
            item, day = self.ask_todo(todo)
            if day:
                todo.date = day
            todo.summary = item
        elif s in ('2', keys[1]):
            day = self.ask_day()
            if type(day) == date:
                todo.date = day
        elif s in ('3', keys[2]):
            todo.date = None
        else:
            self.draw_todo()
            return
        self.todo.sort()
        self.ui.clear_todo()
        self.ui.draw_todo()

    def add_event(self):
        # the footer will be useful, clear it
        self.ui.clear_footer()
        if self.todohl > 0:
            self.add_todo()
            return
        sd = self.ask_starttime()
        if sd == "CUL_CANCEL":
            return  # cancel add event
        ed = self.ask_endtime(sd)
        if ed == "CUL_CANCEL":
            return  # cancel add event
        duration = (ed - sd).total_seconds()
        desc, place = self.ask_description()
        if desc == "CUL_CANCEL":
            return  # cancel add event
        # create event
        e = Event(sd, duration, desc, place)
        self.cal.add_event(e)
        self.cal.sort()
        self.update_calweek(False)
        calday = self.calweek[self.day.isoweekday() - 1]
        self.ui.draw_week()  # redraw all week for multidays event

    def add_todo(self):
        summ, date = self.ask_todo()
        if summ == "CUL_CANCEL":
            return  # canceled new item
        item = Event(date, None, summary=summ, tag=0)
        self.todo.add_event(item)
        self.todo.sort()
        self.ui.clear_todo()
        self.ui.draw_todo()

    def clicked(self):
        x, y = self.ui.get_mouse_position()
        # the todo list has been clicked
        if self.conf.todo and x >= self.ui.xmax:
            if not self.todohl:
                self.toggle_TODO()
            if (y - 2) in range(len(self.todo.events)):
                # select a todo, starting on line 2, todohl starts at 1
                self.todohl = y - 2 + 1
                self.ui.draw_todo()
        elif x <= self.ui.XOFFSET:
            # clicked on the hour? Why?
            pass
        else:  # clicked on the main culendar
            if self.todohl:  # todo list is highlighted
                self.todohl = 0
                self.ui.draw_todo()  # not anymore
            # select the current day
            day_lines = self.ui.lines[0]
            iday = 0
            for i in range(len(day_lines)):
                if x > day_lines[i]:
                    iday += 1
            # Monday is 1
            cur_iday = self.day.isoweekday()
            self.addday(iday - cur_iday)
            # have we clicked on an event?
            calday = self.calweek[iday - 1]
            # Monday is 1, first value is 0
            for i, e in enumerate(calday.events):
                if y in e.vlines and x in e.hlines:
                    self.ehl = i
            self.ui.draw_day()

###############################################################################
#       ask questions
###############################################################################

    def ask_generic(self, question, default=""):
        self.ui.clear_footer()
        self.ui.text_in_footer(question[:self.ui.width], '')
        data = Input(self.ui, default)
        r = data.get_input()
        while r == -1:
            self.ui.update()  # we got a KEY_RESIZE while getting input
            self.ui.clear_footer()
            self.ui.text_in_footer(question[:self.ui.width], '')
            data.screen_update()
            r = data.get_input()
        self.ui.clear_footer()
        if r == 0:
            return data
        else:  # cancel the question
            return "CUL_CANCEL"

    def ask_starttime(self):
        question = _("Enter start time ([hh:mm], [hhmm], [hh] or [h]):")
        start_time = None
        while start_time is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                start_time = data.check_hour(self.day, start=True)
            else:
                data = "CUL_CANCEL"
                start_time = 0
        return self.post_check_input(data, start_time)

    def ask_endtime(self, start_date):
        # ask endtime, return duration of event
        question = _("Enter end time ([hh:mm], [hhmm], [hh] or [h]) or duration ([+mm]):")
        end_date = None
        while end_date is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                end_date = data.check_duration(self.day, start_date)
            else:
                data = "CUL_CANCEL"
                end_date = 0
        return self.post_check_input(data, end_date)

    def ask_description(self, event=None):
        question = _("Enter description[@location]:")
        if event:
            data = self.ask_generic(question, event.summary)
        else:
            data = self.ask_generic(question)
        try:
            desc = data.text
        except Exception:
            return "CUL_CANCEL", ""  # empty place
        # is there a location in the description?
        if "@" in desc:
            place = desc[desc.find("@") + 1:]
            desc = desc[:desc.find("@")]
        else:
            place = ""
        return desc, place

    def ask_place(self, event):
        question = _("Enter location:")
        data = self.ask_generic(question, event.place)
        try:
            return data.text
        except Exception:
            return "CUL_CANCEL"

    def ask_day(self):
        question = _("Enter day ([yyyy/]mm/dd] or [yyyy]mmdd, [Enter for today]):")
        new_day = None
        while new_day is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                new_day = data.check_day(self.day)
            else:
                data = "CUL_CANCEL"
                new_day = 0
        return self.post_check_input(data, new_day)

    def ask_filename(self, mode, question):
        filename = None
        while filename is None:
            data = self.ask_generic(question)
            if data != "CUL_CANCEL":
                filename = data.check_filename(mode)
            else:
                data = "CUL_CANCEL"
                filename = 0
        return self.post_check_input(data, filename)

    def tag_event(self):
        if self.todohl > 0:
            return  # we're in the todolist, nothing to do
        self.ui.clear_footer()
        # select highlighted event if existing
        iday = self.day.isoweekday() - 1
        if len(self.calweek[iday].events) > self.ehl:
            event = self.calweek[iday].events[self.ehl].original
            # the footer will be useful, clear it
            self.ui.clear_footer()
            # ask question
            self.ui.text_in_footer(
                _("Enter tag number [0-7] or [a-z]:"),
                _("(Current tag: ") + str(event.tag) + ")",
            )
            prekey = self.ui.getkey()
            key = prekey[0]
            if key.isdigit() and 0 <= int(key) <= 7:  # local event
                event.tag = int(key)
                self.ui.clear_footer()
            elif key.isalpha():  # caldav event
                if key in self.culdav._tags:
                    idx = self.culdav._tags.index(key)
                    if self.culdav._iscaldav[idx]:  # caldav
                        caldav = self.culdav._icals[idx]
                        try:  # remove from local events
                            self.cal.del_event(event)
                        except Exception:  # it's a switch from another caldav
                            # not a local event; duplicate
                            e = Event(event.date, event.duration,
                                      event.summary, event.place,
                                      event.tag, event.url,
                                      event.caldav, event.colour)
                            event = e
                        # switch from local to caldav local culendars
                        event.tag = key
                        event.colour = self.culdav._list[idx]["colour"]
                        self.culdav._cal.add_event(event)
                        # properly add to the extern caldav
                        self.culdav.add_event(event, caldav)
                        # inform the switch
                        msg = _("Event moved to ")
                        url = caldav.url.url_raw
                        self.ui.inform(msg + url)
                    else:  # it's a webcal, read-only
                        msg = _("Error: can't add an event to a webcal")
                        self.ui.text_in_footer(msg, _("Press any key to continue"))
                        key = self.ui.getkey()
                    self.ui.clear_footer()
                else:
                    errmsg = _(" is not an exisiting calDAV tag")
                    self.ui.inform(key + errmsg)
            else:
                self.ui.clear_footer()
                self.ui.inform(_("Incorrect tag"))
            # redraw with new colour
            self.update_calweek(False)
            self.ui.draw_week()  # week for multidays events

    def copy_event(self, repetitions=1):
        if self.todohl > 0:
            return  # we're in the todolist, nothing to do
        iday = self.day.isoweekday() - 1
        if len(self.calweek[iday].events) > self.ehl:
            # only if an event is selected
            if repetitions > 1:
                question = _("Copy to next [d]ays, [w]eeks, [m]onths or [y]ears?")
            else:
                question = _("Copy to next [d]ay, [w]eek, [m]onth or [y]ear?")
            self.ui.inform(question)
            key = self.ui.getkey()
            if key not in ["d", "D", "w", "W", "m", "M", "y", "Y"]:
                errmsg = _("Invalid answer -- canceled")
                self.ui.inform(errmsg)
                return
            event = self.calweek[iday].events[self.ehl].original
            date = event.date
            duration = event.duration
            summary = event.summary
            place = event.place
            tag = event.tag
            colour = event.colour
            for ind in range(repetitions):
                if key in ["d", "D"]:
                    new_date = date + timedelta(days=ind + 1)
                if key in ["w", "W"]:
                    new_date = date + timedelta(weeks=ind + 1)
                elif key in ["m", "M"]:
                    years = (date.month + ind) // 12
                    month = date.month + ind + 1 - 12 * years
                    try:
                        new_date = date.replace(year=date.year + years, month=month)
                    except Exception:
                        new_date = None
                elif key in ["y", "Y"]:
                    new_date = date.replace(year=date.year + ind + 1)
                if new_date:
                    e = Event(new_date, duration, summary, place, tag, colour=colour)
                    # deal with tags
                    if type(tag) == int:  # local event
                        self.cal.add_event(e)
                        msg = _("Event copied")
                    elif tag in self.culdav._tags:  # caldav event
                        idx = self.culdav._tags.index(tag)
                        caldav = self.culdav._icals[idx]
                        if self.culdav._iscaldav[idx]:  # caldav
                            self.culdav._cal.add_event(e)
                            # properly add to the extern caldav
                            self.culdav.add_event(e, caldav)
                            msg = _("Caldav event remotely copied")
                        else:  # webcal
                            self.cal.add_event(e)
                            msg = _("Webcal event locally copied")
            self.cal.sort()
            if key in ["d", "D"]:
                self.ui.draw_week()  # cause the current week changed
            self.ui.inform(msg)
        # no else: do nothing without an selected event

    def shift_event(self, hours):
        if self.todohl > 0:
            return  # we're in the todolist, nothing to do
        # select highlighted event if existing
        iday = self.day.isoweekday() - 1
        if len(self.calweek[iday].events) > self.ehl:
            event = self.calweek[iday].events[self.ehl].original
            # we can't edit webcal events
            if event.caldav == "webcal":
                # webcals are read-only
                # msg = _("Error: can't change webcal event")
                return  # exit here
            sd = event.date + timedelta(hours=hours)
            event.date = sd
            if event.caldav:  # update the corresponding caldav
                self.culdav.del_event(event)
                self.culdav.add_event(event, event.caldav)
            # give focus to moved event
            self.day = sd.date()
            self.update_calweek(False)
            # search for the event of the new day to highlight the moved one
            iday = self.day.isoweekday() - 1
            for i, e in enumerate(self.calweek[iday].events):
                if e.summary == event.summary and e.place == event.place:
                    self.ehl = i
            self.ui.draw_week()

    def ask_todo(self, todo=None):
        question = _("Enter Todo[@date] ([yyyy/]mm/dd] or [yyyy]mmdd):")
        if todo:
            data = self.ask_generic(question, todo.summary)
        else:
            data = self.ask_generic(question)
        if data == "CUL_CANCEL":
            return "CUL_CANCEL"
        try:
            desc = data.text
            day = None
        except Exception:
            return "CUL_CANCEL", None  # None date
        if "@" in desc:  # is there a date?
            date = desc[desc.find("@") + 1:]
            item = desc[:desc.find("@")]
            data.text = date
            day = data.check_day(self.day)
        else:
            item = desc
        if not type(day) == date:
            day = None
        return item, day

    def post_check_input(self, data, processed_data):
        if data == "KEY_RESIZE" or processed_data == "KEY_RESIZE":
            self.ui.update()
            return "CUL_CANCEL"
        if data == "CUL_CANCEL" or processed_data == "CUL_CANCEL":
            self.ui.clear_footer()
            return "CUL_CANCEL"
        else:
            return processed_data

###############################################################################
# save/import/export
###############################################################################

    def save(self):
        self.cal.save_apts(self.conf, self.culdav)
        self.todo.save_todo(self.conf)

    def importcal(self):
        question = _("Enter the file name to import (ical or calcurse): (filename [tag])")
        filename = self.ask_filename("rb", question)
        tag = 0  # default value
        if filename == "CUL_CANCEL":
            return
        if filename[-2] == " " and filename[-1] in [str(i) for i in range(8)]:
            tag = int(filename[-1])
            filename = filename[:-2]
        try:  # is it an ical?
            cal = Agenda.import_ical(filename, self.conf.categories, tag)
        except Exception:
            try:  # is it from calcurse?
                cal = Agenda.import_calcurse(filename, tag)
            except Exception:
                self.ui.text_in_footer(_("Unable to load file"), _("Press any key to continue"))
                self.ui.getkey()
                self.ui.clear_footer()
                return
        # check for already present event
        # 1) create simple list from self.cal
        existingevents = []
        for se in self.cal.events:
            existingevents.append(se.date.strftime('%Y%m%d%H%M') + str(se.duration) + se.summary.replace(' ', ''))
        for e in cal.events:
            currentevent = (e.date.strftime('%Y%m%d%H%M') + str(e.duration) + e.summary.replace(' ', ''))
            if currentevent not in existingevents:
                self.cal.add_event(e)
        self.cal.sort()
        self.ui.draw_cal()
        self.ui.inform(_("New events imported!"))

    def exportcal(self):
        # returns 1 if successful
        self.ui.clear_footer()
        key = ""
        while key not in ["i", "I", "c", "C", "q", "Q", ""]:
            self.ui.text_in_footer(_("Export to [i]cal or to [c]alcurse?"), "[i/c]")
            key = self.ui.getkey()
            if key in ["q", "Q", ""]:
                self.ui.clear_footer()
                return
        question = _("Enter the file name to export:")
        filename = self.ask_filename("wb", question)
        if filename == "CUL_CANCEL":
            return
        if key in ["c", "C"]:
            self.cal.export_calcurse(filename)
        else:
            self.cal.export_ical(filename, self.conf.categories)
        self.ui.inform(_("Calendar exported!"))



import importlib.resources
from gettext import translation
from locale import (
    LC_ALL,
    Error,
    setlocale,
)
from os import environ
from pathlib import Path
from sys import (
    platform,
    stderr,
)
from typing import (
    Callable,
    cast,
)

DOMAIN = 'culendar'


def get_default_lang() -> str:
    """Inspired from gajim source code"""
    if platform == 'win32':
        from ctypes import windll  # type: ignore[attr-defined]
        from locale import windows_locale
        user_default_ui_language = windll.kernel32.GetUserDefaultUILanguage()
        return windows_locale[user_default_ui_language]
    elif platform == 'darwin':
        from AppKit import NSLocale  # type: ignore
        return NSLocale.currentLocale().languageCode()
    else:
        from locale import getdefaultlocale
        return getdefaultlocale()[0] or 'en'


def get_translation_func() -> Callable:
    try:
        setlocale(LC_ALL, '')
    except Error as error:
        print(error, file=stderr)
    LANG = get_default_lang()
    if platform == 'win32':
        # Set the env var on Windows because gettext.find() uses it to find the translation
        # Use LANGUAGE instead of LANG to keep settings like LC_TIME
        environ['LANGUAGE'] = LANG
    package_dir = cast(Path, importlib.resources.files(DOMAIN))
    locale_dir = package_dir / 'locale'
    try:
        _translation = translation(DOMAIN, locale_dir)
        _gettext = _translation.gettext
        try:
            from locale import bindtextdomain as locale_bindtextdomain
            locale_bindtextdomain(DOMAIN, locale_dir)
        except ImportError:
            pass
    except OSError:
        from gettext import NullTranslations
        _translation = NullTranslations()  # type: ignore[assignment]
        _gettext = _translation.gettext
        print('No translations found for', LANG, file=stderr)
    return _gettext


_ = get_translation_func()

__all__ = ['_']



import curses
from datetime import (
    date,
    timedelta,
)

from .i18n import _


class UI:
    X_MAX_WITH_MOUSE = 223  # limit to clickable without bug
    XOFFSET = 6  # left offset to draw hours
    YOFFSETTOP = 3  # top offset to draw the header
    YOFFSETBOT = 4  # bottom offset to draw the footer

    def __init__(self, culendar, conf, screen):
        self.ctrl = culendar
        self.conf = conf
        self.screen = screen
        self.xmax = 0
        self.ymax = 0
        self.day_size = 0
        self.todocol = 0
        self.width = 0
        self.hour_size = 0
        self.lines = [[], []]
        self.do_update_on_resize = True
        self.update_screen_props()

    @property
    def hmin(self):
        return self.conf.hmin

    @property
    def hmax(self):
        return self.conf.hmax

    @property
    def daynb(self):
        return 7 if self.conf.WE else 5

    def update(self):
        self.update_screen_props()
        # try:
        # everything changes, clear screen and redraw stuff
        self.clear_cal()
        self.draw_cal()
        # FIXME only catch proper exception or test for minimal terminal size instead
        # except Exception as e:
        #     # culendar couldn't draw it: it's ridiculously small terminal
        #     self.dead_duck(e)

    def update_screen_props(self):
        self.ymax, self.xmax = self.screen.getmaxyx()
        if self.conf.mouse:
            self.xmax = min(self.xmax, self.X_MAX_WITH_MOUSE)
        if self.conf.todo:  # keep space for a todo
            xmax = self.xmax - round(self.xmax * self.conf.todowidth)
            self.day_size = (xmax - self.XOFFSET) // self.daynb
            # update xmax to be hour place + a multiple of day_size
            xmax = 7 + self.day_size * self.daynb
            self.todocol = self.xmax - xmax
            self.xmax = xmax
        else:
            self.day_size = (self.xmax - self.XOFFSET) // self.daynb
            self.todocol = 0
        self.width = self.xmax + self.todocol
        self.hour_size = (self.ymax - self.YOFFSETTOP - self.YOFFSETBOT) // (self.hmax - self.hmin)
        self.linepos()

    def linepos(self):
        # locations of [day lines, hour lines]
        self.lines = [[], []]
        self.lines[0].append(self.XOFFSET)
        for id in range(self.daynb - 1):
            self.lines[0].append(self.lines[0][id] + self.day_size)
        # the last one is in the last position
        self.lines[0].append(self.xmax - 1)
        self.lines[1].append(self.YOFFSETTOP)
        for ih in range(self.hmax - self.hmin):
            self.lines[1].append(self.lines[1][ih] + self.hour_size)

    def clear_cal(self):
        self.screen.clear()

    def draw_cal(self):
        if self.conf.mouse:
            self.draw_mousedeath()
        self.draw_table()
        self.draw_hours()
        self.draw_week()
        if self.conf.todo:
            self.draw_todo()

    def draw_mousedeath(self):
        ymax, real_xmax = self.screen.getmaxyx()
        if self.conf.todo:
            xmax = self.xmax + self.todocol
        else:
            xmax = self.xmax
        if real_xmax > xmax:
            msg = _(" ☠ DON'T CLICK HERE ☠ ")
            nbmsg = (real_xmax - xmax) // len(msg)
            remain = (real_xmax - xmax) - len(msg) * nbmsg
            msg = msg * nbmsg + msg[:remain]
            for y in range(self.ymax - 1):
                self.screen.addstr(y, xmax, msg, curses.A_REVERSE)
            self.screen.addstr(self.ymax - 1, xmax, msg[:-1], curses.A_REVERSE)

    def draw_table(self):
        # draw vertical lines to bottom point
        bottom = self.lines[1][-1] - 2
        self.screen.vline(2, 0, curses.ACS_VLINE, bottom)
        for ld in self.lines[0]:
            self.screen.vline(2, ld, curses.ACS_VLINE, bottom)
        # draw horizontal lines up to the end
        self.screen.hline(1, 1, curses.ACS_HLINE, self.xmax - 2)
        for lh in self.lines[1]:
            self.screen.hline(lh, 1, curses.ACS_HLINE, self.xmax - 2)
        for lh in self.lines[1]:
            for ld in self.lines[0]:
                # draw crosses
                self.screen.addch(lh, ld, curses.ACS_PLUS)
            # draw left and right tees
            self.screen.addch(lh, 0, curses.ACS_LTEE)
            self.screen.addch(lh, self.lines[0][-1], curses.ACS_RTEE)
        for ld in self.lines[0]:
            # draw top and bottom tees
            self.screen.addch(1, ld, curses.ACS_TTEE)
            self.screen.addch(self.lines[1][-1], ld, curses.ACS_BTEE)
        # draw corners
        self.screen.addch(1, 0, curses.ACS_ULCORNER)
        self.screen.addch(1, self.lines[0][-1], curses.ACS_URCORNER)
        self.screen.addch(self.lines[1][-1], 0, curses.ACS_LLCORNER)
        self.screen.addch(self.lines[1][-1], self.lines[0][-1], curses.ACS_LRCORNER)

    def draw_hours(self):
        self.screen.addstr(2, 1, _('hours'))
        for i, h in enumerate(range(self.hmin, self.hmax)):
            if h < 10:
                self.screen.addstr(self.lines[1][i], 3, f'{h}h')
            else:
                self.screen.addstr(self.lines[1][i], 2, f'{h}h')

    def draw_week(self):
        self.ctrl.update_calweek(False)
        self.draw_header()
        self.draw_daynames()  # redraw day names
        if self.conf.debug:
            for iday in range(self.daynb):
                self.draw_day(iday)
            self.draw_hl()
        else:
            # try:
            for iday in range(self.daynb):
                self.draw_day(iday)
            self.draw_hl()
            # FIXME only catch proper exception or test for minimal terminal size instead
            # except Exception as e:  # couldn't draw it: too small terminal
            #     self.dead_duck(e)

    def draw_header(self):
        # erase previous header
        self.screen.hline(0, 0, " ", self.xmax - 1)
        if self.ctrl.todohl:
            style = curses.A_NORMAL
        else:
            style = curses.A_BOLD
        w = (
            _(' Week ') + str(self.ctrl.day.isocalendar()[1]) + ', '
            + self.ctrl.day.strftime("%B") + ' ' + str(self.ctrl.day.isocalendar()[0])
        )
        self.screen.addstr(0, (self.xmax - len(w)) // 2 + 1, w, style)
        self.screen.addstr(0, 0, str(self.conf.keys['help'][0]), curses.A_BOLD)
        self.screen.addstr(0, 1, _(": help"), curses.A_BOLD)

    def draw_daynames(self):
        # first day is monday = 1
        curday = date.fromordinal(self.ctrl.day.toordinal() - self.ctrl.day.isoweekday() + 1)
        for id in range(self.daynb):
            d = curday.strftime("%a %d/%m")
            pos = self.lines[0][id] + (self.day_size - len(d)) // 2 + 1
            if curday == self.ctrl.day:
                self.screen.addstr(2, pos, d, curses.A_BOLD)
            else:
                self.screen.addstr(2, pos, d)
            # increment day
            curday = date.fromordinal(curday.toordinal() + 1)

    def draw_day(self, iday=-1, hl=0):
        self.clear_day(iday)
        if iday == -1:  # defaults to current day
            iday = self.ctrl.day.isoweekday() - 1
            hl = 1  # not in draw_week, we'll draw the highlight
        calday = self.ctrl.calweek[iday]
        for e in calday.events:
            self.draw_event(e, curses.A_NORMAL)
        if hl:
            self.draw_hl()

    def clear_day(self, iday=-1):
        if iday == -1:  # defaults to current day
            iday = self.ctrl.day.isoweekday() - 1
        if iday == self.daynb - 1:
            width = self.lines[0][-1] - self.lines[0][-2] - 1
        else:
            width = self.day_size - 1
        # day_size - 1: doesn't include vertical separating lines
        for v in range(self.lines[1][0], self.lines[1][-1] + 1):
            # +1 to be sure to include last line
            self.screen.hline(v, self.lines[0][iday] + 1, " ", width)
            if v in self.lines[1]:
                # if it's a horizontal line, redraw it
                self.screen.hline(v, self.lines[0][iday] + 1, curses.ACS_HLINE, width)

    def draw_event(self, e, flag):
        # flag is bold if highlighted, normal if not
        try:
            colour = self.conf.colours[e.tag] + curses.A_REVERSE + flag
        except Exception:  # alphabetic tag, that's a caldav, use e.color
            colour = self.conf.colours[e.colour] + curses.A_REVERSE + flag
        normal = curses.A_REVERSE + flag
        if type(e.hlines) == tuple:  # cf calendary.py compute_hlines()
            dduck = [
                _(r"\_x<  too   >x_/")[0:e.hlines[1]],
                _(r"\_x<  many  >x_/")[0:e.hlines[1]],
                _(r"\_x< events >x_/")[0:e.hlines[1]],
            ]
            for iv, v in enumerate(e.vlines):
                self.screen.addstr(v, e.hlines[0], dduck[iv % 3])
            return
        width = len(e.hlines)
        # draw background
        for v in e.vlines:
            if self.conf.colourset == 0:
                self.screen.hline(v, e.hlines[0], " ", width, colour)
            elif self.conf.colourset == 1:
                self.screen.addstr(v, e.hlines[0], " ", colour)
                self.screen.addstr(v, e.hlines[0] + width - 1, " ", colour)
                self.screen.hline(v, e.hlines[0] + 1, " ", width - 2, normal)
            elif self.conf.colourset == 2:
                self.screen.hline(v, e.hlines[0], " ", width, normal)
            else:
                self.screen.hline(v, e.hlines[0], " ", width, normal)
        # prepare the text
        t = e.summary
        if len(t) > width:
            t = t[0:width]
        # draw text for all cases, bold for highlighted event
        posx = e.hlines[0] + (width - len(t)) // 2
        posy = e.vlines[0] + (e.vlines[-1] - e.vlines[0]) // 2
        if self.conf.colourset == 0:
            self.screen.addstr(posy, posx, t, colour)
        elif self.conf.colourset == 1:
            self.screen.addstr(posy, posx, t, normal)
            if len(t) == width:
                self.screen.addstr(posy, posx, t[0], colour)
                self.screen.addstr(posy, posx + width - 1, t[-1], colour)
        elif self.conf.colourset == 2:
            self.screen.addstr(posy, posx, t, colour)
        else:
            self.screen.addstr(posy, posx, t, normal)
        # if there is a location, add it if it's a two lines event
        if e.place != "" and e.vlines[-1] > e.vlines[0]:
            # prepare the text
            p = e.place
            if len(p) > width:
                p = p[0:width]
            posx = e.hlines[0] + (width - len(p)) // 2
            if self.conf.colourset == 0:
                self.screen.addstr(posy + 1, posx, p, colour)
            elif self.conf.colourset == 1:
                self.screen.addstr(posy + 1, posx, p, normal)
                if len(p) == width:
                    self.screen.addstr(posy + 1, posx, p[0], colour)
                    self.screen.addstr(posy + 1, posx + width - 1, p[-1], colour)
            elif self.conf.colourset == 2:
                self.screen.addstr(posy + 1, posx, p, colour)
            else:
                self.screen.addstr(posy + 1, posx, p, normal)

    def draw_hl(self):
        if self.conf.todo and self.ctrl.todohl > 0:
            self.draw_todo()  # erase previous hl, draw new one
            return
        # else: normal event hl
        self.clear_hl()
        self.draw_daynames()
        iday = self.ctrl.day.isoweekday() - 1
        maxhl = len(self.ctrl.calweek[iday].events)
        # is there anything to highlight?
        if self.ctrl.ehl < maxhl and maxhl > 0:
            self.draw_event(self.ctrl.calweek[iday].events[self.ctrl.ehl], curses.A_BOLD)
            self.draw_footer(self.ctrl.calweek[iday].events[self.ctrl.ehl])

    def clear_hl(self):
        self.clear_footer()
        # we toggled the WE off; nothing to do
        if self.ctrl.is_hl_hidden_WE:
            pass
        # if we change the hl of current day
        elif self.ctrl.is_hl_same_day:
            # just redraw in normal font
            iday = self.ctrl.day.isoweekday() - 1
            maxhl = len(self.ctrl.calweek[iday].events)
            # is there anything to highlight?
            if self.ctrl.prevehl < maxhl and maxhl > 0:
                self.draw_event(self.ctrl.calweek[iday].events[self.ctrl.prevehl], curses.A_NORMAL)
        # if we change from less than a week
        elif self.ctrl.is_hl_same_week:
            # redraw previous day
            calday = self.ctrl.calweek[self.ctrl.previous_day.isoweekday() - 1]
            for e in calday.events:
                self.draw_event(e, curses.A_NORMAL)
        # else: nothing to do, new week clears everything

    def clear_footer(self):
        for y in range(max(0, self.ymax - 3), self.ymax):
            self.screen.hline(y, 0, " ", self.width)

    def text_in_footer(self, *messages: list[str]):
        if len(messages) > 3:
            raise ValueError("footer text can only takes up to 3 message lines")
        for i, msg in enumerate(messages[::-1], 1):
            self.screen.addstr(self.ymax - i, 0, msg)

    def draw_footer(self, e):
        enddate = e.original.date + timedelta(seconds=e.original.duration)
        if e.place:
            desc = e.summary + " @ " + e.place
        else:
            desc = e.summary
        text = _('Summary: ') + desc
        self.text_in_footer(
            _('Begin date: {}').format(e.original.date),
            _('End date:   {}').format(enddate),
            text[:self.width - 1],
        )

    def draw_todo(self):
        if self.ctrl.todohl:
            style = curses.A_BOLD
        else:
            style = curses.A_NORMAL
        title = _("TODO")  # noqa: T101
        xpos = self.xmax + (self.todocol - len(title)) // 2
        self.screen.addch(0, self.xmax - 1, curses.ACS_VLINE)
        self.screen.addch(1, self.xmax - 1, curses.ACS_PLUS)
        self.screen.addch(self.lines[1][-1], self.xmax - 1, curses.ACS_RTEE)
        self.screen.vline(self.lines[1][-1] + 1, self.xmax - 1, curses.ACS_VLINE, self.ymax - self.lines[1][-1] - 4)
        self.screen.hline(1, self.xmax, curses.ACS_HLINE, self.todocol)
        self.screen.addstr(0, xpos, title, style)
        for idx, item in enumerate(self.ctrl.todo.events):
            if idx + 1 == self.ctrl.todohl:
                style = curses.A_BOLD
            else:
                style = curses.A_NORMAL
            if item.date:
                date = item.date.strftime("%d/%m")
                self.screen.addstr(2 + idx, self.xmax + self.todocol - len(date), date, style)
            if len(item.summary) < self.todocol:
                # + " " to ensure a space before the date
                text = item.summary + " "
            else:
                text = item.summary[:self.todocol - 1]
            self.screen.addstr(2 + idx, self.xmax + 1, text, style)
        if self.ctrl.todohl > 0 and len(self.ctrl.todo.events) > 0:
            self.draw_todohl()

    def draw_todohl(self):
        """draw highlighted todo in the footer"""
        self.clear_footer()
        if len(self.ctrl.todo.events) + 1 > self.ctrl.todohl:
            item = self.ctrl.todo.events[self.ctrl.todohl - 1]
            date = ""
            if item.date:
                date = (
                    " -- " + str(item.date.day)
                    + "/" + str(item.date.month)
                    + "/" + str(item.date.year)
                )
            text = item.summary + date
            self.text_in_footer(text[:self.width - 1])

    def clear_todo(self):
        for y in range(2, self.ymax):
            self.screen.hline(y, self.xmax, " ", self.todocol)

    def dead_duck(self, exception: Exception):
        text = _(r"\_x< A dead duck. That's all such a small terminal deserves.")
        if self.conf.debug:
            curses.endwin()
            print(text)
        else:
            self.clear_cal()
            self.screen.addstr(0, 0, text[0:(self.width - 1) * (self.ymax - 1)])
            key = ""
            while key not in self.conf.keys['quit'] and key != 'KEY_RESIZE':
                key = self.getkey()
            if key in self.conf.keys['quit']:
                if self.conf.debug:
                    print(exception)
                exit()
            else:
                self.update()

    def getkey(self):
        """
        function to deal properly with all kind of typed key
        get_wch() returns a str for standard keys;
        for functions keys, sometimes string names, sometimes escaped sequences,
        sometimes ordinal integers.
        stupid function that translates different possibilities into strings
        for the fun, it's terminal-dependant

        roughly tested with urvxt, screen, xterm terminals
        probably doesn't work with yours
        """
        key = self.screen.get_wch()
        if key == "":
            # half delay mode of 1/10s: no input key raises an error
            # assumption: no humain being type ESC + real keys so fast
            curses.halfdelay(1)
            try:
                # get the false keys and stack it
                while 1:
                    key += str(self.screen.get_wch())
            except Exception:
                # the bundle of special keys stopped:
                # back to normal mode
                curses.cbreak()
                if key in ["[b", "[1;2B"]:
                    key = "kDN"  # shift down
                elif key in ["[a", "[1;2A"]:
                    key = "kUP"  # shift right
                elif key in ["[d", "[1;2D"]:
                    key = "KEY_SLEFT"
                elif key in ["[c", "[1;2C"]:
                    key = "KEY_SRIGHT"
                # non curses character: control+key
                elif key in ["Ob", "[1;5B"]:
                    key = "KEY_CDOWN"
                elif key in ["Oa", "[1;5A"]:
                    key = "KEY_CUP"
                elif key in ["Od", "[1;5D"]:
                    key = "KEY_CLEFT"
                elif key in ["Oc", "[1;5C"]:
                    key = "KEY_CRIGHT"
                elif key in ["OP"]:
                    key = "KEY_F(1)"
                elif key in ["OQ"]:
                    key = "KEY_F(2)"
                elif key in ["OR"]:
                    key = "KEY_F(3)"
                elif key in ["OS"]:
                    key = "KEY_F(4)"
                elif key in ["[1~", "OH", "[H"]:
                    key = "KEY_HOME"
                elif key in ["[4~", "OF", "[F"]:
                    key = "KEY_END"
                else:
                    if not self.conf.debug and len(key) > 1:
                        key = ""
        if type(key) == int:
            if key in [410]:
                key = "KEY_RESIZE"
            elif key in [258]:
                key = "KEY_DOWN"
            elif key in [259]:
                key = "KEY_UP"
            elif key in [260]:
                key = "KEY_LEFT"
            elif key in [261]:
                key = "KEY_RIGHT"
            elif key in [262, 362]:
                key = "KEY_HOME"
            elif key in [360, 385]:
                key = "KEY_END"
            elif key in [263]:
                key = "KEY_BACKSPACE"
            elif key in [265]:
                key = "KEY_F(1)"
            elif key in [266]:
                key = "KEY_F(2)"
            elif key in [267]:
                key = "KEY_F(3)"
            elif key in [268]:
                key = "KEY_F(4)"
            elif key in [269]:
                key = "KEY_F(5)"
            elif key in [270]:
                key = "KEY_F(6)"
            elif key in [271]:
                key = "KEY_F(7)"
            elif key in [272]:
                key = "KEY_F(8)"
            elif key in [273]:
                key = "KEY_F(9)"
            elif key in [274]:
                key = "KEY_F(10)"
            elif key in [275]:
                key = "KEY_F(11)"
            elif key in [276]:
                key = "KEY_F(12)"
            elif key in [330]:
                key = "KEY_DC"
            elif key in [331]:
                key = "KEY_IC"
            elif key in [338]:
                key = "KEY_NPAGE"
            elif key in [339]:
                key = "KEY_PPAGE"
            elif key in [353]:
                key = "KEY_BTAB"
            elif key in [513, 336]:
                key = "kDN"
            elif key in [529, 337]:
                key = "kUP"
            elif key in [393]:
                key = "KEY_SLEFT"
            elif key in [402]:
                key = "KEY_SRIGHT"
            # non curses character: control+key
            elif key in [514, 525]:
                key = "KEY_CDOWN"
            elif key in [530, 566]:
                key = "KEY_CUP"
            elif key in [523, 545]:
                key = "KEY_CLEFT"
            elif key in [528, 560]:
                key = "KEY_CRIGHT"
            elif key in [409]:
                key = "KEY_MOUSE"
            else:
                if not self.conf.debug:
                    key = ""
                else:
                    key = str(key)
        if key == "KEY_RESIZE" and self.do_update_on_resize:
            self.update()
        return key

    def get_mouse_position(self):
        try:
            _, x, y, _, _ = curses.getmouse()
        except Exception:
            x = -1
            y = -1
        return x, y

    def inform(self, *messages: list[str]):
        """use footer to give feedback"""
        self.clear_footer()
        self.text_in_footer(*[
            msg[0:self.width - 1] for msg in messages
        ])
        self.screen.refresh()

    def debug(self, elt):
        self.screen.addstr(self.ymax - 2, 50, 'debug:   {}'.format(elt))
        self.screen.getch()  # just a pause
        self.screen.hline(self.ymax - 2, 1, " ", self.width - 2)



from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from os import (
    path,
    rename,
)

from icalendar import Calendar
from icalendar import Event as IEvent


class Event:
    """
    Event class
    """

    def __init__(self, date, duration, summary, place="", tag=0,
                 url=None, caldav=None, colour=None):
        self._date = date                 # datetime.datetime()
        self._duration = int(duration)    # seconds
        self._summary = summary.strip()   # text
        self._place = place.strip()       # text
        self._tag = tag
        self._colour = colour  # default colour is tag; useful for caldav
        self._samehour = 0     # number of simultaneous events
        self._hlines = None    # horizontal lines of event
        self._vlines = None    # vertical lines of event
        # CalDAV
        self._url = url
        self._caldav = caldav

    @property
    def original(self) -> 'Event':
        return self

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, d):
        self._date = d

    @property
    def duration(self):
        return self._duration

    @duration.setter
    def duration(self, d):
        self._duration = d

    @property
    def summary(self):
        return self._summary

    @summary.setter
    def summary(self, s):
        self._summary = s

    @property
    def place(self):
        return self._place

    @place.setter
    def place(self, p):
        self._place = p

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, t):
        self._tag = t

    @property
    def samehour(self):
        return self._samehour

    @samehour.setter
    def samehour(self, samehour):
        self._samehour = samehour

    @property
    def hlines(self):
        return self._hlines

    @hlines.setter
    def hlines(self, hlines):
        self._hlines = hlines

    @property
    def vlines(self):
        return self._vlines

    @vlines.setter
    def vlines(self, vlines):
        self._vlines = vlines

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, url):
        self._url = url

    @property
    def caldav(self):
        return self._caldav

    @caldav.setter
    def caldav(self, caldav):
        self._caldav = caldav

    @property
    def colour(self):
        return self._colour

    @colour.setter
    def colour(self, colour):
        self._colour = colour

    def compute_vlines(self, ui):
        start_hour = self.date.hour
        start_min = self.date.minute
        early_start = 0
        late_end = 0
        end_time = start_hour * 3600 + start_min * 60 + self.duration
        # if starting hour doesn't fit, make it fit
        if start_hour < ui.hmin:
            start_hour = ui.hmin
            start_min = 0
            early_start = 1
        if end_time < start_hour * 3600:
            end_time = start_hour * 3600
        if end_time/3600 > ui.hmax:
            late_end = 1
        if start_hour >= ui.hmax:
            start_hour = ui.hmax
            start_min = 0
            late_end = 0  # already late
        # don't overflow
        end_time = min(ui.hmax * 3600, end_time)
        end_hour = int(end_time // 3600)
        end_min = int((end_time - end_hour * 3600) // 60)
        # time to lines
        # select the lines
        startline = start_hour - ui.hmin
        endline = end_hour - ui.hmin
        shline = ui.lines[1][startline] + 1  # 1st line after hour line
        ehline = ui.lines[1][endline] + 1    # same as start hour
        # lines to range of line
        if ui.hour_size == 1:
            # small terminals: starts on line event
            shline = shline - 1
            ehline = ehline - 1
        else:  # tall terminal, work between lines, deal with minutes
            minute_size = round(60 / (ui.hour_size - 1))
            if ui.hour_size > 2:
                # change start line only for wide hours
                shline += start_min // minute_size
            if end_min == 0:
                ehline -= 1  # stops before hour line
            else:
                ehline += end_min // minute_size
            # special case out of schedule
            if start_hour == ui.hmax:
                shline -= 1
        # failproof
        if ehline <= shline:
            ehline = shline + 1
        # overlaps the table when outside the limits
        shline -= early_start
        ehline += late_end
        # finally define the range
        self.vlines = range(shline, ehline)

    def compute_hlines(self, ui, col, width, maxevents, table):
        start_vline = table[self.vlines[0]]
        e_width = width // maxevents
        left = 0
        while sum(start_vline[left:left + e_width]) != 0:
            left += 1  # width//maxevents
            # have we gone too far?
            if left + e_width > len(start_vline):
                # reduce event width
                e_width = len(start_vline) - left
        # first column picked from lines
        for v in self.vlines:
            for i in range(e_width):
                table[v][left+i] = 1
        # first column: right of vertical day line + offset of multievent
        first_col = ui.lines[0][col] + 1 + left
        # finally define the range
        self.hlines = range(first_col, first_col + e_width)
        # too many events or too small terminal
        if len(self.hlines) < 1:
            self.hlines = (ui.lines[0][col] + 1, width)

    def to_icalevent(self, categories=None):
        ie = IEvent()
        ie.add('dtstart', self.date)
        ie.add('duration', timedelta(seconds=self.duration))
        ie.add('summary', self.summary)
        if self.place:
            ie.add('location', self.place)
        if self.tag and type(self.tag) == int:
            if len(categories[self.tag]) > 1:
                ie.add('category', categories[self.tag][1])
            else:
                ie.add('category', self.tag)
        if self.url:
            ie.add('url', self.url)
        return ie

    def split_event(self):
        diffdayfirst = self.date.toordinal()
        last_day = self.date + timedelta(seconds=self.duration)
        diffdaylast = last_day.toordinal()
        if diffdayfirst == diffdaylast:
            return [self]
        else:
            # first subevent
            end_date = self.date.replace(hour=23, minute=59)
            e1 = SubEvent(self, self.date, (end_date-self.date).total_seconds())
            subevents = [e1]
            day_span = diffdaylast-diffdayfirst
            for i in range(1, day_span+1):
                if i != day_span:  # not last event
                    new_day = self.date.replace(hour=0, minute=0) + timedelta(days=i)
                    subevents.append(SubEvent(self, new_day, 86340))  # 23h59 in seconds
                else:
                    new_day = self.date.replace(hour=0, minute=0) + timedelta(days=i)
                    duration = (last_day.hour*60 + last_day.minute)*60
                    subevents.append(SubEvent(self, new_day, duration))
        return subevents


class SubEvent(Event):
    def __init__(self, event: Event, date, duration):
        self._orig_event = event
        super().__init__(
            date,
            duration,
            summary=event.summary,
            place=event.place,
            tag=event.tag,
            url=event.url,
            caldav=event.caldav,
            colour=event.colour,
        )

    @property
    def original(self) -> 'Event':
        return self._orig_event

class Agenda:
    """
    Agenda class
    """
    def __init__(self):
        self._events = []

    def add_event(self, e):
        self._events.append(e)

    def del_event(self, e):
        self._events.remove(e.original)

    @property
    def events(self):
        return self._events

    def sort(self):
        """sort by date and time."""
        maxdate = date(9999, 12, 31)
        self._events = sorted(self._events, key=lambda e: e.date if e.date else maxdate)

    @classmethod
    def from_calendar(cls, calendar, categories, default_tag, caldav=None):
        agenda = cls()
        for ie in calendar.walk('vevent'):
            sdate = ie['dtstart'].dt
            try:  # is there an end date?
                edate = ie['dtend'].dt
                # if so, compute duration
                dur = (edate - sdate).total_seconds()
            except KeyError:
                try:  # there is thus a duration
                    dur = ie['duration'].dt.total_seconds()
                except KeyError:  # no end date, no duration… Are you kidding me?
                    dur = 0  # it exists, that's all
            try:  # is there a summary?
                summ = ie['summary'].lstrip().replace('\n', "-")
            except KeyError:  # if none, NULL summary
                summ = " "
            try:  # is there a location?
                place = ie['location'].lstrip().replace('\n', "-")
            except KeyError:  # if none, NULL summary
                place = ""
            try:  # is there a category?
                cat = ie['category']
                tag = default_tag
                if type(cat) == list:  # several categories in a single event
                    for c in cat:
                        for t in range(1, 8):
                            if c.lstrip() in categories[t]:
                                # if category exists, tag it
                                tag = t
                else:  # single category
                    for t in range(1, 8):
                        if cat.lstrip() in categories[t]:
                            # if category exists, tag it
                            tag = t
            except KeyError:  # no category
                tag = default_tag
            # special case for events without hour:
            # set 1st hour to 0:00 and duration to 24h
            if type(sdate) == date:
                sdate = datetime.combine(sdate, time(0))
                dur = 24 * 60 * 60
            # check the potential timezone and put to local
            if sdate.tzinfo:
                sdate = sdate.astimezone()
                sdate = sdate.replace(tzinfo=None)
            if caldav:
                e = Event(sdate, dur, summ, place, tag, caldav=caldav)
            else:
                e = Event(sdate, dur, summ, place, tag)
            agenda.add_event(e)
        return agenda

    @classmethod
    def extract_day(cls, cal, culdav, day):
        # create a list of subcals for each day of the week
        calday = cls()
        for e in cal.events + culdav.calweek.events:
            # if the right day, add the event
            if e.date.toordinal() == day.toordinal():
                calday.add_event(e)
        calday.sort()
        return calday

    @classmethod
    def extract_week(cls, cal, culdav, day, ui, autosync=True):
        # create a list of subcals for each day of the week
        calweek = []
        for i in range(7):
            calweek.append(cls())
        # selects the events in caldavs, if required
        if autosync:
            culdav.sync()
        # first day of the current week; +1 for starting on monday in ordinal
        firstday = day.toordinal() - day.isoweekday() + 1
        for e in cal.events + culdav.events:
            diffdayfirst = e.date.toordinal() - firstday
            diffdaylast = (e.date + timedelta(seconds=e.duration)).toordinal() - firstday
            dayspan = set(range(diffdayfirst, diffdaylast+1))
            # if in the right week, we have
            #if 0 <= diffday <= 6:
            if dayspan.intersection(set(range(7))):  # if the event days interect the week
                if diffdayfirst == diffdaylast:
                    calweek[diffdayfirst].add_event(e)
                else:
                    for sub_e in e.split_event():
                        diffday = sub_e.date.toordinal() - firstday
                        if 0 <= diffday <= 6:
                            calweek[diffday].add_event(sub_e)
        for iday in range(ui.daynb):  # for all printed days
            calweek[iday].sort()
            for e in calweek[iday].events:
                # compute each vertical position for events
                e.compute_vlines(ui)
            # initialize no event on all lines
            nbevent = [0 for h in range(ui.ymax)]
            for e in calweek[iday].events:
                for i in e.vlines:
                    nbevent[i] += 1
            maxevent = max(nbevent)
            # if last idayumn, the day size is not ui.day_size
            if iday == ui.daynb - 1:
                width = ui.lines[0][-1] - ui.lines[0][-2] - 1
            else:
                width = ui.day_size - 1
            # day_size - 1: doesn't include vertical separating lines
            # table of position of events on day
            table = [[0 for d in range(width)] for y in range(ui.ymax)]
            for e in calweek[iday].events:
                maxevent = max([nbevent[i] for i in e.vlines])
                e.compute_hlines(ui, iday, width, maxevent, table)
        return calweek

    @classmethod
    def import_ical(cls, filename, categories, default_tag=0):
        with open(filename, "r") as f:
            calendar = Calendar.from_ical(f.read())
        return cls.from_calendar(calendar, categories, default_tag)

    @classmethod
    def import_calcurse(cls, filename, t=0):
        cal = cls()
        with open(filename) as f:
            for line in f:
                smon = int(line[0:2])
                sday = int(line[3:5])
                syea = int(line[6:10])
                shou = int(line[13:15])
                smin = int(line[16:18])
                emon = int(line[22:24])
                eday = int(line[25:27])
                eyea = int(line[28:32])
                ehou = int(line[35:37])
                emin = int(line[38:40])
                summ = line[42:-1]  # strips final \n
                sdate = datetime(syea, smon, sday, shou, smin)
                edate = datetime(eyea, emon, eday, ehou, emin)
                e = Event(sdate, (edate-sdate).total_seconds(), summ, tag=t)
                cal.add_event(e)
        return cal

    def export_ical(self, filename, categories):
        ical = Calendar()
        ical.add('prodid', '-//From Culendar')
        ical.add('version', '2.0')
        for e in self.events:
            ie = e.to_icalevent(categories)
            ical.add_component(ie)
        with open(filename, "wb") as f:
            f.write(ical.to_ical())

    def export_calcurse(self, filename):
        lines = ""
        for e in self.events:
            line = str(e.date.strftime('%m/%d/%Y @ %H:%M'))
            line += " -> "
            enddate = e.date + timedelta(seconds=e.duration)
            line += str(enddate.strftime('%m/%d/%Y @ %H:%M'))
            line += " |"
            line += e.summary
            if e.place:
                line += "@"
                line += e.place
            line += "\n"
            lines += line
        with open(filename, "w") as f:
            f.write(lines)

    @classmethod
    def load(cls, filename, todofilename):
        return cls.load_apts(filename), cls.load_todo(todofilename)

    @classmethod
    def load_apts(cls, filename, caldav=None):
        cal = cls()
        with open(filename) as f:
            for line in f:
                syea = int(line[0:4])
                smon = int(line[5:7])
                sday = int(line[8:10])
                shou = int(line[13:15])
                smin = int(line[16:18])
                eyea = int(line[22:26])
                emon = int(line[27:29])
                eday = int(line[30:32])
                ehou = int(line[35:37])
                emin = int(line[38:40])
                rema = line[42:-1]  # strips final \n
                # search for a tag
                try:
                    if "|" in rema:
                        tag = int(rema[rema.find("|") + 1:])
                        rema = rema[:rema.find("|")]
                    else:
                        tag = 0
                except ValueError:
                    tag = 0
                # search for a location delimiter
                if "@" in rema:
                    summ = rema[:rema.find("@")]
                    place = rema[rema.find("@") + 1:]
                else:
                    summ = rema
                    place = ""
                sdate = datetime(syea, smon, sday, shou, smin)
                edate = datetime(eyea, emon, eday, ehou, emin)
                e = Event(sdate, (edate-sdate).total_seconds(), summ, place, tag, caldav=caldav)
                cal.add_event(e)
        cal.sort()
        return cal

    @classmethod
    def load_todo(cls, todofilename):
        # Todo list: ugly hack, use Agenda
        # Todo item: Event with no duration and optional date
        todo = cls()
        with open(todofilename) as f:
            for line in f:
                try:  # if there is a date
                    syea = int(line[0:4])
                    smon = int(line[5:7])
                    sday = int(line[8:10])
                    sdate = date(syea, smon, sday)
                    rema = line[14:-1]  # strips final \n
                except ValueError:
                    rema = line[:-1]  # strips final \n
                    sdate = None
                try:
                    # if no |, empty tag and rema doesn't change
                    tag = int(rema[rema.find("|") + 1:])
                    summ = rema[:rema.find("|")]
                except ValueError:
                    tag = 0
                    summ = rema
                e = Event(sdate, None, summary=summ, tag=tag)
                todo.add_event(e)
        todo.sort()
        return todo

    def cal2txt(self, colour=None):
        lines = ""
        for e in self.events:
            line = str(e.date.strftime('%Y/%m/%d @ %H:%M'))
            line += " -> "
            enddate = e.date + timedelta(seconds=e.duration)
            line += str(enddate.strftime('%Y/%m/%d @ %H:%M'))
            line += " |"
            line += e.summary
            if e.place:
                line += "@"
                line += e.place
            if colour:
                line += "|"
                line += str(colour)
            elif e.tag:
                line += "|"
                line += str(e.tag)
            line += "\n"
            lines += line
        return lines

    def todo2txt(self):
        lines = ""
        for e in self.events:
            if e.date:
                line = str(e.date.strftime('%Y/%m/%d'))
                line += " -> "
            else:
                line = ""
            line += e.summary
            if e.tag:
                line += "|"
                line += str(e.tag)
            line += "\n"
            lines += line
        return lines

    def save_apts(self, conf, culdav=None):
        # main Agenda
        # backup current one
        if path.exists(conf.datafile):
            rename(conf.datafile, conf.datafile + '.old')
        with open(conf.datafile, "w") as f:
            f.write(self.cal2txt())
        if culdav:
            culdav.save_local()

    def save_todo(self, conf):
        # Todo file
        # backup current one
        if path.exists(conf.todofile):
            rename(conf.todofile, conf.todofile + '.old')
        with open(conf.todofile, "w") as f:
            f.write(self.todo2txt())



#!/usr/bin/env python3
from curses import wrapper

from .culendar import Culendar


def main(stdscr, *args):
    Culendar(stdscr).run_loop()


def run(args=None):
    """here starts culendar!"""
    if args is None:
        from sys import argv as args
    wrapper(main, *args)


if __name__ == "__main__":
    run()



from datetime import datetime
from os import path
from urllib import request

from caldav import DAVClient
from caldav.lib.error import DAVError
from icalendar import Calendar

from .calendar import Agenda
from .i18n import _


class Culdav:
    """Culendar CalDav"""
    def __init__(self, caldav_dict, datapath, ui):
        self._list = caldav_dict
        self._datapath = datapath
        self.ui = ui
        # list of Icalendar, for each caldav
        # proper url if webcal; None if offline
        self._icals = []
        self._iscaldav = []     # list of True/False
        self._tags = []         # associated tags
        self._cal = Agenda()    # single list of all events
        self._cals = []         # list of event lists, by caldav

    @property
    def icals(self):
        return self._icals

    @property
    def events(self):
        return self._cal.events

    def initialize(self):
        for cdav_idx, cdav in enumerate(self._list):
            self.initialize_single(cdav, cdav_idx)

    def initialize_single(self, cdav, cdav_idx):
        try:
            if "@" in cdav["url"]:  # there is a password in the url
                head_user_password, tail = cdav["url"].split("@")
                head, user_password = head_user_password.split("://")
                user = user_password.split(":")[0]
                url = head + "://" + user + ":" + "*****" + "@" + tail
            else:
                url = cdav["url"]
            msg = _("Loading caldav ") + str(cdav_idx + 1) + _(": ") + url
            self.ui.inform(msg)
            if cdav["url"][:4] == "http":  # caldav
                flag = True  # it's a proper caldav
                dav_client = self.get_client(cdav)
                self.add_calendar(dav_client)
            if cdav["url"][:6] == "webcal":  # webcal
                flag = False  # it's a webcal
                if cdav["url"][:7] == "webcals":  # webcal
                    urlh = "https" + cdav["url"][7:]
                else:
                    urlh = "http" + cdav["url"][6:]
                self.read_webcal_url(urlh)
                self._icals.append(urlh)
            self._iscaldav.append(flag)
        except Exception as e:
            err = _("Error: can't contact caldav ") + cdav["url"]
            if self.load_local(cdav["url"], cdav["username"], check=True):
                err += _(" -- using offline copy")
            err += _(" -- press any key to continue")
            self.ui.inform(err, str(e))
            self.ui.getkey()
            # append None to keep same size as _list
            self._icals.append(None)
            self._iscaldav.append(False)
        self.add_tag(cdav)

    def get_client(self, cdav):
        if len(cdav["username"]) != 0:
            client = DAVClient(cdav["url"], username=cdav["username"], password=cdav["password"])
        else:
            client = DAVClient(cdav["url"])
        return client

    def add_calendar(self, dav_client):
        principal = dav_client.principal()
        calendars = principal.calendars()
        if len(calendars) == 0:  # useful for my poor test of python-radicale
            cal = principal.make_calendar()
        else:
            try:
                cal = dav_client.calendar(url=dav_client.url)
            except DAVError:
                raise
            try:
                cal.get_supported_components()  # ensure calendar exist
            except (KeyError, DAVError) as e:
                # guess the correct calendar
		        uuid=dav_client.url.url_raw.split("/")[-2]
        		got_uuid=False
        		for c in calendars:
        			if uuid in str(c):
		    	    	cal = c
			    	    got_uuid=True
            			break
        		if not got_uuid:
		        	cal = calendars[0]
                try:
                    cal.get_supported_components()  # ensure calendar exist
                except (KeyError, DAVError):
                    # nothing left to be done, re-raise first error
                    raise e
        self._icals.append(cal)

    def add_tag(self, cdav):
        if cdav:
            self._tags.append(cdav["tag"])
        else:
            self._tags.append(None)

    def read_webcal_url(self, url):
        if "@" not in url:  # no password
            return request.urlopen(url).read().decode()
        else:
            # keep between http[s] *://* user:pwd *@* url
            user_and_password = url.split("@")[0].split("://")[1]
            user = user_and_password.split(":")[0]
            # if there are ":" in the password, join them again
            password = ":".join(user_and_password.split(":")[1:])
            # remove login and password from url
            url_clean = "".join(url.split(user + ":" + password + "@"))
            # create a password manager and an opener
            pwd_mgr = request.HTTPPasswordMgrWithDefaultRealm()
            pwd_mgr.add_password(None, url_clean, user, password)
            handler = request.HTTPBasicAuthHandler(pwd_mgr)
            opener = request.build_opener(handler)
            # finally read and decode the data
            return opener.open(url_clean).read().decode()

    def sync(self):
        # extracts all the events of the the ICalendar
        # set all events in a single culendar/Agenda class
        self.ui.inform(_("Synchronising caldavs…"))
        # resets the differents cals
        self._cals = []
        for ind, cal in enumerate(self._icals):
            tmpcal = Agenda()
            if self._iscaldav[ind]:
                ical = cal.events()
                for e in ical:  # a list of ical events, each one in a calendar
                    agendacul = Agenda.from_calendar(Calendar.from_ical(e.data), [], self._tags[ind], cal)
                    ecul = agendacul.events[0]  # a single event
                    ecul.url = e.canonical_url
                    ecul.colour = self._list[ind]["colour"]
                    tmpcal.add_event(ecul)
            else:
                try:
                    # if unreachable, cal was None and thus, error
                    cal = self.read_webcal_url(cal)
                except Exception:
                    pass
                # reached and a webcal
                if cal and cal[:15] == "BEGIN:VCALENDAR":
                    # in a webcal, _icals is url
                    # in a webcal, we get everything.
                    agendacul = Agenda.from_calendar(Calendar.from_ical(cal), [], self._tags[ind], "webcal")
                    for e in agendacul.events:
                        e.colour = self._list[ind]["colour"]
                        tmpcal.add_event(e)
                else:  # unreachable, use offline copy if any
                    url = self._list[ind]["url"]
                    username = self._list[ind]["username"]
                    tmpcal = self.load_local(url, username)
                    if cal:  # we got something, but not a wecal
                        warn = _("Warning: contacting ")
                        if len(url) > 20:
                            warn += url[:20] + "[…]"
                        else:
                            warn += url
                        warn += _(" gave a ")
                        warn += cal[:23]
                        warn += _(" -- using offline copy")
                        warn += _(" -- press any key to continue")
                        self.ui.inform(warn)
                        self.ui.getkey()
            if tmpcal:  # avoid None from loading offline apts
                tmpcal.sort()  # sort for better offline text copy
                self._cals.append(tmpcal)
        self._cal = Agenda()
        for cal in self._cals:
            for e in cal.events:
                self._cal.add_event(e)
        self.save_local()

    def del_caldav(self, list_idx):
        # _list already removed in confscreen
        # clean what is required:
        # from highest position to lowest to pop easily
        list_idx.sort(reverse=True)
        for idx in list_idx:
            _ = self._tags.pop(idx)
            self._icals.pop(idx)
            self._iscaldav.pop(idx)

    def update(self, oldconf):
        list_idx = []
        for idx, conf in enumerate(oldconf):
            if conf not in self._list:
                list_idx.append(idx)
        if list_idx:
            self.del_caldav(list_idx)
        idx = 0
        for conf in self._list:
            if conf not in oldconf:
                self.initialize_single(conf, idx)
                idx += 1
        self.sync()

    def save_local(self):
        # Caldav local saves - no backup, it's online
        if not self._icals:
            return
        for cdav_idx, cdav in enumerate(self._list):
            if self._icals[cdav_idx]:  # is there anything to save
                url = cdav["url"]
                suffix = url.replace("/", "-") + cdav["username"]
                fname = "apts." + suffix
                with open(self._datapath+"/"+fname, "w") as f:
                    f.write(self._cals[cdav_idx].cal2txt(cdav["colour"]))

    def load_local(self, url, username, check=False):
        suffix = url.replace("/", "-") + username
        filename = self._datapath + "/" + "apts." + suffix
        if path.exists(filename):
            if check:  # just check existence
                return True
            else:
                return Agenda.load_apts(filename, caldav="webcal")
        else:  # no local copy exists
            return False if check else None

    def add_event(self, event, cdav):
        event.caldav = cdav
        cal = Calendar()
        icalevent = event.to_icalevent()
        # caldav reguires a UID
        icalevent.add("uid", hash(datetime.now()))
        cal.add_component(icalevent)
        # https://caldav.readthedocs.io/en/latest/caldav/objects.html#caldav.objects.Calendar.add_event
        tmpevent = cdav.add_event(cal.to_ical().decode())
        event.url = tmpevent.canonical_url

    def del_event(self, event):
        # https://caldav.readthedocs.io/en/latest/caldav/objects.html#caldav.objects.Calendar.event_by_url
        event.caldav.event_by_url(event.url).delete()
        # delete from local cal in order to avoid syncing caldav to update events
        self._cal.del_event(event.original)
